'''
    常用的 HTTP 请求——封装了发生错误时重试的功能
'''

import logging
from os import makedirs, path
from time import sleep, time

import requests

from .constants import HEADERS, STATUS_CODES, TIMEOUT

# 设置日志
logging.basicConfig(
    level=logging.WARNING,
    format='%(levelname)s %(message)s'
)

# 禁用SLL证书警告
requests.urllib3.disable_warnings()


# 发生错误时再次请求
def retry_request(method, url, retries=3, **kwargs):
    retry = 1
    kwargs.update({
        'headers': kwargs.get('headers', HEADERS),
        'timeout': kwargs.get('timeout', TIMEOUT),
        'verify': kwargs.get('verify', False),
    })
    while True:
        try:
            resp = requests.request(method, url, **kwargs)
            if resp.status_code in STATUS_CODES:
                return resp
            if retry >= retries:
                logging.warning(f"[{resp.status_code}]: {url}")
                return False
            retry += 1
        except Exception as error:
            if retry >= retries:
                logging.error(f"{error}: {url}")
                return False
            retry += 1
            sleep(3)  # 发生错误时休眠若干秒


def retry_get(url, retries=3, **kwargs):
    return retry_request('GET', url, retries=retries, **kwargs)


def retry_post(url, data=None, retries=3, **kwargs):
    return retry_request('POST', url, data=data, retries=retries, **kwargs)


# 下载文件
def download(src, dst, retries=8, **kwargs):
    resp = retry_get(src, retries, **kwargs)
    if resp:
        with open(dst, 'wb') as f:
            f.write(resp.content)
        logging.info(src)


# 下载文件且以时间戳命名文件
def download_file(src, folder='.'):
    '''Download the file and name the file with a timestamp.'''
    if not path.isdir(folder):
        makedirs(folder)
    ext = path.splitext(src)[-1]
    dst = path.join(folder, f'{int(time())}{ext}')
    download(src, dst)
    return dst


# 将请求数据写入文件
def save_html(response, fn='default.html'):
    if not fn.endswith('.html'):
        fn = '%s.html' % fn
    with open(fn, 'wb') as fp:
        fp.write(response.content)



'''HTML parsing helper function.'''

import re

from .constants import USELESS_TAGS
from .quick import replace_ws


# 清除空格符号，然后替换连续的空白符号
def rm_ws(string, char='\n'):
    '''Clear a space character, then 
    replace the continuous blank symbols.'''
    return re.sub(r'\s+', char, replace_ws(string))


# 移除指定标签及其内容，默认同时移除其他标签
# 但保留其内容，最后清除或替换多余空白符
def rm_tags(html, char='', tags=USELESS_TAGS, default=True):
    '''Remove the specified label and its content, 
    while removing other labels but retaining their content.'''
    regex = '|'.join([r'<%s\b.*?</%s>|<%s\s*/>' %
                      (tag, tag, tag) for tag in tags])
    if default:
        regex += r'|</?([^ >/]+).*?>'
    retags = re.compile(regex, re.DOTALL | re.IGNORECASE)
    string = retags.sub('', html)
    return rm_ws(string, char)



'''Building a modular crawler template system based on Jinja2'''

import re
import warnings
from datetime import datetime

from jinja2 import Environment, FileSystemLoader

from .utils import (replace_all, replace_ws, retry_get, retry_post, rm_tags,
                    rm_ws)
from .utils.constants import DATETIME_12, WHITESPACE


class ModularLoader(FileSystemLoader):

    def __init__(self, searchpath, encoding="utf-8", followlinks=False):
        if not isinstance(searchpath, (tuple, set, list)):
            searchpath = [searchpath]
        self.searchpath = searchpath
        self.encoding = encoding
        self.followlinks = followlinks
        self.templates = []
        self.flag = False
        self.head, self.body, self.foot = '', '', ''

    def get_source(self, environment, template):
        x, y, z = super().get_source(environment, template)
        if self.flag:
            return x, y, z
        self.flag = True
        return self.head+self.body+x+self.foot, y, z

    def extends(self, templates):
        if not isinstance(templates, (tuple, set, list)):
            templates = (templates,)
        ext = ''
        for template in templates:
            ext += '{% extends "template" %}\n'.replace('template', template)
        self.head = ext + self.head

    def include(self, templates, left=False):
        if not isinstance(templates, (tuple, set, list)):
            templates = (templates,)
        ext = ''
        for template in templates:
            ext += '{% include "template" %}\n'.replace('template', template)
        if left:
            self.body += ext
        else:
            self.foot += ext

    def include_left(self, templates):
        self.include(templates, left=True)

    def add_block(self, block):
        self.body += '{%- block bk -%}\n'.replace('bk', block)
        self.foot = '{%- endblock bk -%}\n'.replace('bk', block) + self.foot

    def reset(self):
        self.flag = False
        self.head, self.body, self.foot = '', '', ''


def render_templatefile(dst, templatefile, **kwargs):
    default_loader = FileSystemLoader(
        kwargs.pop('templates_folder', './templates'))
    env = Environment(loader=kwargs.pop('loader', default_loader))
    template = env.get_template(templatefile)
    with open(dst, 'w', encoding='utf-8') as fp:
        fp.write(template.render(**kwargs))


# 可根据实际需求修改部分代码
def genspider(home_url, templatefile, dst, spider, **kwargs):
    response = retry_get(home_url)
    if not response:
        raise Exception('Request failed, invalid URL or parameter.')

    try:
        string = response.content.decode('utf8')
    except UnicodeDecodeError:
        string = response.content.decode('gbk')

    source = re.findall(r'(?is)<title>(.*?)</title>', string)[0]

    # 清理标题中的干扰字符
    clear_chars = kwargs.pop('clear_chars', WHITESPACE)
    source = re.sub(r'\s', '', replace_all(source, clear_chars + WHITESPACE))

    # 仅供参考，建议自行修改
    kwargs.update({
        'datetime': datetime.now().strftime(DATETIME_12),
        'spider': spider,
        'source': source,
        # 'region': replace_all(source, list(kwargs.get('city', '')))[:3] if spider[5] != '0' else '',
        'home_url': home_url,
        'author': kwargs.pop('author', 'White Turing'),
    })

    spiderfile = f'{dst}/{spider}.py'
    render_templatefile(spiderfile, templatefile, **kwargs)



'''Parsing HTML forms.'''

import re
import warnings

from lxml import html

from .htmlib import rm_ws
from .httplib import retry_get
from .quick import contains


# 判断表格是否包含相关数据
def is_table(tables, within, without):
    '''Determine whether the table contains relevant data.'''
    is_tab, table = False, None
    for table in tables:
        string = re.sub(r'\s', '', table.text_content())
        if contains(string, within, without):
            is_tab = True
            break
    return is_tab, table


# 提取表格全部文本
def extract_string(table) -> str:
    '''Extract all text in the table.'''
    return rm_ws(table.text_content())


# 按行按列提取表格文本放入元组列表
def extract_list(table) -> list:
    '''Extract the table text by row and column 
    and put it into the tuple list.'''
    rows = table.xpath('tr | */tr')
    form = []
    for row in rows:
        cols = row.xpath('td | th')
        res = []
        for col in cols:
            res.append(rm_ws(col.text_content(), ''))
        if res:
            form.append(tuple(res))
    return form


# 判断是否是每行列数相同的表格
def is_equal(form):
    '''Determine whether it is a table with the 
    same number of columns in each row.'''
    return all(len(form[0]) == len(
        form[pos]) for pos in range(1, len(form)))


# 打包全部操作
def parse_form(form_url, within, without=tuple(), encoding='utf-8', sep='</t', to_list=False):
    response = retry_get(form_url)
    if not response:
        raise Exception('Request failed, invalid URL or parameter.')

    string = response.content.decode(encoding)
    elements = html.fromstring(string.replace(sep, '\n'+sep))

    tables = elements.xpath('//table')  # 找到全部表格
    is_tab, table = is_table(tables, within, without)
    if is_tab:
        return extract_list(table) if to_list else extract_string(table)

    warnings.warn('Cannot find table with required data.')
    return False



'''常用操作的函数封装'''

from functools import partial

from .constants import WHITESPACE


# 判断是否包含某些元素且不包含某些元素
def contains(string, within, without=tuple()) -> bool:
    '''Determine if it contains certain elements
       and does not contain certain elements.'''
    if any(k in string for k in without):
        return False
    if any(k in string for k in within):
        return True
    return False


# 替换一组字符
def replace_all(string, words=tuple(), char='') -> str:
    '''Replace a set of characters.'''
    for w in words:
        string = string.replace(w, char)
    return string


# 清除空格符号
replace_ws = partial(replace_all, words=WHITESPACE)


# 将路径分割为上一级路径、文件名和扩展名
def split_path(path):
    '''Split the path/url into upper-level path/url, file name, and extension.'''
    import os.path
    ph, filext = os.path.split(path)  # 分割为上一级路径和包含后缀的文件名
    fil, ext = os.path.splitext(filext)  # 分割为文件名和后缀
    return ph, fil, ext


# 偶然发现了另一种方法
def split_path2(path):
    '''Split the path/url into upper-level path/url, file name, and extension.'''
    from pathlib import Path
    ph = Path(path)
    return ph.parent, ph.stem, ph.suffix



'''Building a modular crawler template system based on Jinja2'''

# Note: To upload this project, you must:

'''
python setup.py sdist
pip install dist/spider-renderer-0.2.1.tar.gz
python setup.py bdist_wheel
pip install twine
twine upload dist/*
'''

import os.path

from setuptools import setup

from shutil import rmtree
if os.path.exists('dist'):
    rmtree('dist')

# What packages are required for this module to be executed?
requires = [
    'requests',
    'jinja2',
]

# Import the README and use it as the long-description.
cwd = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(cwd, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='spider-renderer',
    packages=['renderer', 'renderer.utils'], # `renderer.utils`: `renderer/utils`
    version='0.2.1',
    license='Apache 2.0',
    author='White Turing',
    author_email='fu.jiawei@outlook.com',
    description='Building a modular crawler template system based on Jinja2.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/ffujiawei/spider-renderer',
    keywords=['spider', 'renderer', 'scrapy'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.6.0',
    install_requires=requires,
)
